import asyncio
from threading import Thread

from websockets.legacy.server import WebSocketServerProtocol, serve


class WebSocketServer:
    """
    Async WebSocket server for live reloading.
    Notifies clients when they should reload.
    """

    host: str
    port: int
    clients: set[WebSocketServerProtocol]
    loop: asyncio.AbstractEventLoop | None
    thread: Thread | None

    def __init__(self, host: str = "localhost", port: int = 8765):
        self.host = host
        self.port = port
        self.clients = set()
        self.loop = None
        self.thread = None

    async def _handler(self, ws: WebSocketServerProtocol):
        """Handle incoming connections by adding to client set."""
        self.clients.add(ws)
        try:
            await ws.wait_closed()
        finally:
            self.clients.remove(ws)

    def start(self):
        """Spin up server."""

        def run():
            # set up async event loop
            self.loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self.loop)
            # start server
            ws_server = serve(
                ws_handler=self._handler,
                host=self.host,
                port=self.port,
            )
            # add server to event loop
            self.loop.run_until_complete(ws_server)
            self.loop.run_forever()

        # spawn async serever in a thread
        self.thread = Thread(target=run, daemon=True)
        self.thread.start()

    async def _broadcast(self, message: str):
        """Broadcast message to all connected clients."""
        for ws in self.clients.copy():
            try:
                await ws.send(message)
            except Exception:
                self.clients.discard(ws)

    def notify_all(self, message: str = "reload"):
        """Notify all connected clients."""
        if self.loop and self.clients:
            asyncio.run_coroutine_threadsafe(
                coro=self._broadcast(message), loop=self.loop
            )



from datetime import date
from pathlib import Path

import pytest

from zona.builder import build, discover, split_metadata
from zona.metadata import Metadata


def test_split_metadata(tmp_path: Path):
    content = """---
title: Test Post
date: 2025-06-03
description: This is a test.
---

# Hello World
    """
    test_file = tmp_path / "test.md"
    test_file.write_text(content)
    meta, content = split_metadata(test_file)

    assert isinstance(meta, Metadata)
    assert meta.title == "Test Post"
    assert meta.description == "This is a test."
    assert meta.date == date(2025, 6, 3)
    assert content == "# Hello World"


def test_no_metadata(tmp_path: Path):
    content = "# Hello World"
    test_file = tmp_path / "hello-world.md"
    test_file.write_text(content)
    meta, content = split_metadata(test_file)

    assert isinstance(meta, Metadata)
    assert meta.title == "Hello World"
    assert meta.description is None
    assert meta.date == date.today()
    assert content == "# Hello World"


def test_malformed_metadata(tmp_path: Path):
    with pytest.raises(ValueError):
        tests = {
            """
---
title: Test Post
date: not a date!!!
description: This is a test.
---
        """,
            """
---
title: Test Post
foobar:
    something: what???
description: This is a test.
---
        """,
            """
---
title Test Post
description: This is a test.
---
        """,
        }
        for i, content in enumerate(tests):
            test_file = tmp_path / str(i)
            test_file.write_text(content)
            split_metadata(test_file)


def test_discover(tmp_path: Path):
    contentd = tmp_path / "content"
    staticd = contentd / "static"
    templatesd = tmp_path / "templates"
    outd = tmp_path / "out"

    for d in [contentd, staticd, templatesd, outd]:
        d.mkdir()
    md_file = contentd / "post.md"
    md_file.write_text("""---
title: Test Post
date: 2025-06-03
description: This is a test.
---

# Hello World
    """)

    style = staticd / "style.css"
    style_content = """
p {
  color: red;
  text-align: center;
}
"""
    style.write_text(style_content)

    items = discover(tmp_path, outd)

    assert len(items) == 2

    md_item = items[0]
    assert md_item.source == md_file
    assert md_item.destination.name == "index.html"
    assert md_item.destination.is_relative_to(outd)
    assert md_item.url == "post.md"
    assert isinstance(md_item.metadata, Metadata)
    assert md_item.metadata.title == "Test Post"
    assert md_item.content is not None
    assert md_item.content.strip() == "# Hello World"

    st_item = items[1]
    assert st_item.source == style
    assert st_item.destination.name == "style.css"
    assert st_item.destination.is_relative_to(outd)
    assert st_item.url == "static/style.css"
    assert st_item.metadata is None


def test_build(tmp_path: Path):
    contentd = tmp_path / "content"
    staticd = contentd / "static"
    templatesd = tmp_path / "templates"
    outd = tmp_path / "out"

    for d in [contentd, staticd, templatesd, outd]:
        d.mkdir()
    md_file = contentd / "post.md"
    md_content = """---
title: Test Post
date: 2025-06-03
description: This is a test.
---

# Hello World
    """
    md_file.write_text(md_content)

    style = staticd / "style.css"
    style_content = """
p {
  color: red;
  text-align: center;
}
"""
    style.write_text(style_content)

    items = discover(tmp_path, outd)
    build(tmp_path, items)
    html = (outd / "post" / "index.html").read_text()
    assert html.strip() == "<h1>Hello World</h1>"
    s = (outd / "static" / "style.css").read_text()
    assert s.strip() == style_content.strip()



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime, tzinfo
from pathlib import Path
from typing import Any
from zoneinfo import ZoneInfo

import yaml
from dacite import Config as DaciteConfig
from dacite import from_dict

from zona.log import get_logger

logger = get_logger()


def find_config(start: Path | None = None) -> Path | None:
    logger.debug("Searching for config file...")
    current = (start or Path.cwd()).resolve()

    for parent in [current, *current.parents]:
        candidate = parent / "config.yml"
        if candidate.is_file():
            logger.debug(f"Config file {candidate} found.")
            return candidate
    logger.debug("Couldn't find config file.")
    return None


SitemapConfig = dict[str, str]


@dataclass
class PostDefaultsConfig:
    description: str = "A blog post"


@dataclass
class BlogConfig:
    dir: str = "blog"
    defaults: PostDefaultsConfig = field(
        default_factory=PostDefaultsConfig
    )


@dataclass
class HighlightingConfig:
    enabled: bool = True
    theme: str = "ashen"
    wrap: bool = False


@dataclass
class LinksConfig:
    external_new_tab: bool = True


@dataclass
class MarkdownConfig:
    image_labels: bool = True
    tab_length: int = 2
    syntax_highlighting: HighlightingConfig = field(
        default_factory=HighlightingConfig
    )
    links: LinksConfig = field(default_factory=LinksConfig)


@dataclass
class BuildConfig:
    clean_output_dir: bool = True
    include_drafts: bool = False


@dataclass
class ReloadConfig:
    enabled: bool = True
    scroll_tolerance: int = 100


@dataclass
class ServerConfig:
    reload: ReloadConfig = field(default_factory=ReloadConfig)


@dataclass
class AuthorConfig:
    name: str = "John Doe"
    email: str = "john@doe.net"


@dataclass
class FeedConfig:
    enabled: bool = True
    timezone: tzinfo = field(default_factory=lambda: ZoneInfo("UTC"))
    path: str = "rss.xml"
    link: str = "https://example.com"
    title: str = "Zona Website"
    description: str = "My zona website."
    language: str = "en"
    author: AuthorConfig = field(default_factory=AuthorConfig)


IGNORELIST = [".marksman.toml"]


def parse_timezone(s: Any) -> tzinfo:
    if isinstance(s, str):
        return ZoneInfo(s)
    else:
        raise TypeError(
            f"Expected {str}, got {type(s)} for config key timezone"
        )


@dataclass
class ZonaConfig:
    base_url: str = "/"
    feed: FeedConfig = field(default_factory=FeedConfig)
    # dictionary where key is name, value is url
    sitemap: SitemapConfig = field(
        default_factory=lambda: {"Home": "/"}
    )
    # list of globs relative to content that should be ignored
    ignore: list[str] = field(default_factory=lambda: IGNORELIST)
    markdown: MarkdownConfig = field(default_factory=MarkdownConfig)
    build: BuildConfig = field(default_factory=BuildConfig)
    blog: BlogConfig = field(default_factory=BlogConfig)
    server: ServerConfig = field(default_factory=ServerConfig)

    @classmethod
    def from_file(cls, path: Path) -> ZonaConfig:
        with open(path, "r") as f:
            raw = yaml.safe_load(f)
        config: ZonaConfig = from_dict(
            data_class=cls,
            data=raw,
            config=DaciteConfig(type_hooks={tzinfo: parse_timezone}),
        )
        return config



import logging

from rich.logging import RichHandler

_LOGGER_NAME = "zona"


def setup_logging(
    level: str = "INFO",
):
    logger = logging.getLogger(_LOGGER_NAME)
    logger.setLevel(level.upper())

    logger.propagate = False

    if not logger.handlers:
        handler = RichHandler(rich_tracebacks=True, show_path=False)
        handler.setLevel(level.upper())
        formatter = logging.Formatter("%(message)s")
        handler.setFormatter(formatter)
        logger.addHandler(handler)


def get_logger(name: str = _LOGGER_NAME) -> logging.Logger:
    return logging.getLogger(name)



from importlib.metadata import version as __version__
from pathlib import Path
from typing import Annotated

import typer

from zona import server
from zona._alias_group import AliasGroup
from zona.builder import ZonaBuilder
from zona.layout import initialize_site
from zona.log import get_logger, setup_logging
from zona.templates import (
    list_templates,
    write_all_templates,
    write_named_templates,
)

app = typer.Typer(cls=AliasGroup, add_completion=False)
logger = get_logger()


@app.command("init | setup")
def init(
    root: Annotated[
        Path | None,
        typer.Argument(
            help="Target directory to populate as a Zona project",
        ),
    ] = None,
):
    """
    Initialize a Zona website project.

    The required directory structure is created,
    and the default configuration file is included.

    Optionally specify the ROOT directory.
    """
    logger.info("Initializing site...")
    initialize_site(root)


@app.command("build | make | b")
def build(
    root: Annotated[
        Path | None,
        typer.Argument(
            help="Directory containing config.yml",
        ),
    ] = None,
    output: Annotated[
        Path | None,
        typer.Option(
            "--output", "-o", help="Location to write built website"
        ),
    ] = None,
    draft: Annotated[
        bool,
        typer.Option("--draft", "-d", help="Include drafts."),
    ] = False,
):
    """
    Build the website.

    Optionally specify the ROOT and OUTPUT directories.
    """
    if draft:
        print("Option override: including drafts.")
    builder = ZonaBuilder(
        cli_root=root, cli_output=output, draft=draft
    )
    builder.build()


@app.command("serve | preview")
def serve(
    root: Annotated[
        Path | None,
        typer.Argument(
            help="Directory containing config.yml",
        ),
    ] = None,
    host: Annotated[
        str,
        typer.Option(
            "--host", help="Hostname for live preview server."
        ),
    ] = "localhost",
    port: Annotated[
        int,
        typer.Option(
            "--port",
            "-p",
            help="Port number for live preview server.",
        ),
    ] = 8000,
    output: Annotated[
        Path | None,
        typer.Option(
            "--output",
            "-o",
            help="Location to write built website. Temporary directory by default.",
        ),
    ] = None,
    final: Annotated[
        bool,
        typer.Option("--final", "-f", help="Don't include drafts."),
    ] = False,
    live_reload: Annotated[
        bool | None,
        typer.Option(
            "--live-reload/--no-live-reload",
            "-l/-L",
            help="Automatically reload web preview. Overrides config.",
            show_default=False,
        ),
    ] = None,
):
    """
    Build the website and start a live preview server.

    The website is rebuilt when the source is modified.

    Optionally specify the ROOT and OUTPUT directories.
    """
    if final:
        print("Preview without drafts.")
    else:
        print("Preview with drafts.")
    if live_reload is None:
        reload = None
    else:
        reload = live_reload
    server.serve(
        root=root,
        output=output,
        draft=not final,
        host=host,
        port=port,
        user_reload=reload,
    )


@app.command("templates | temp")
def templates(
    names: Annotated[
        list[str] | None,
        typer.Argument(
            help="List of included templates to copy. Suffix .html is optional.",
            show_default=False,
        ),
    ] = None,
    list_: Annotated[
        bool,
        typer.Option(
            "-l",
            "--list",
            help="List available included templates.",
        ),
    ] = False,
    root: Annotated[
        Path | None,
        typer.Option("--root", help="Override root directory."),
    ] = None,
    all: Annotated[
        bool,
        typer.Option(
            "-a", "--all", help="Write all included templates."
        ),
    ] = False,
    force: Annotated[
        bool,
        typer.Option(
            "-f",
            "--force",
            help="Danger: overwrite existing templates.",
        ),
    ] = False,
):
    """Write copies of default templates included with zona to the user template
    directory. These can be used to override the defaults."""
    if all:
        write_all_templates(root, force=force)
        print("All templates written.")
    elif list_:
        print(list_templates())
    elif not names:
        raise typer.BadParameter(
            "Must provide template names unless --all is specified."
        )
    else:
        write_named_templates(names, root, force=force)
        print(f"Wrote template(s): {', '.join(names)}.")


def version_callback(value: bool):
    if value:
        print(f"Zona version: {__version__('zona')}")
        raise typer.Exit()


@app.callback()
def main_entry(
    version: Annotated[  # pyright: ignore[reportUnusedParameter]
        bool | None,
        typer.Option(
            "--version",
            callback=version_callback,
            is_eager=True,
            help="Print version info and exit.",
        ),
    ] = None,
    verbosity: Annotated[
        str,
        typer.Option(
            "--verbosity",
            "-v",
            help="Logging verbosity. One of INFO, DEBUG, WARN, ERROR.",
        ),
    ] = "info",
) -> None:
    """
    Opinionated static site generator.

    Supply --help after any subcommand for more details.
    """
    setup_logging(verbosity)


def main():
    app()



from pathlib import Path
from typing import Literal

from jinja2 import Environment, FileSystemLoader, select_autoescape

from zona import util
from zona.config import ZonaConfig
from zona.layout import Layout, discover_root
from zona.markdown import md_to_html
from zona.models import Item


def get_header(template_dir: Path) -> str | None:
    md_header = template_dir / "header.md"
    html_header = template_dir / "header.html"
    if md_header.exists():
        return md_to_html(md_header.read_text(), None)
    elif html_header.exists():
        return html_header.read_text()


def get_footer(template_dir: Path) -> str | None:
    md_footer = template_dir / "footer.md"
    html_footer = template_dir / "footer.html"
    if md_footer.exists():
        return md_to_html(md_footer.read_text(), None)
    elif html_footer.exists():
        return html_footer.read_text()


def get_template_map() -> util.TemplateMap:
    """Returns a mapping of template names to ZonaResource tuples."""
    map = {
        str(r).removeprefix("templates/").removesuffix(".html"): r
        for r in util.get_resources("templates")
    }
    return map


def _get_templates(
    cli_root: Path | None,
) -> tuple[Path, list[util.ZonaResource]]:
    root = discover_root(cli_root)
    resources = util.get_resources("templates")
    return root, resources


def write_named_templates(
    names: list[str], cli_root: Path | None, force: bool = False
):
    """Writes a list of template names to the user directory.
    Raises FileExistsError for existing files if force == False."""
    root = discover_root(cli_root)
    map = get_template_map()
    for name in names:
        if name not in map:
            raise KeyError(f"{name} is not a valid template name.")
        map[name].write_with_root(root, force=force)


def write_all_templates(cli_root: Path | None, force: bool = False):
    """Write all internal templates to the user directory.
    Raises FileExistsError for existing files if force == False."""
    root, resources = _get_templates(cli_root)
    for r in resources:
        r.write_with_root(root, force=force)


def list_templates() -> str:
    return "\n".join(get_template_map().keys())


# TODO: add a recent posts element that can be included elsewhere?
class Templater:
    def __init__(
        self,
        config: ZonaConfig,
        template_dir: Path,
        post_list: list[Item],
    ):
        # build temporary template dir
        self.env: Environment = Environment(
            loader=FileSystemLoader(template_dir),
            autoescape=select_autoescape(["html", "xml"]),
        )
        self.config: ZonaConfig = config
        self.template_dir: Path = template_dir
        self.footer: str | None = get_footer(template_dir)
        self.post_list: list[Item] = post_list

    def render_header(self):
        template = self.env.get_template("header.html")
        return template.render(site_map=self.config.sitemap)

    def render_item(self, item: Item, content: str) -> str:
        env = self.env
        meta = item.metadata
        assert meta is not None
        if meta.template is None:
            if item.post:
                template_name = "page.html"
            else:
                template_name = "basic.html"
        else:
            template_name = (
                meta.template
                if meta.template.endswith(".html")
                else meta.template + ".html"
            )
        template = env.get_template(template_name)
        header: str | Literal[False] = (
            self.render_header() if meta.header else False
        )
        footer: str | Literal[False] = (
            self.footer if self.footer and meta.footer else False
        )
        return template.render(
            content=content,
            url=item.url,
            metadata=meta,
            header=header,
            footer=footer,
            is_post=item.post,
            newer=util.normalize_url(item.newer.url)
            if item.newer
            else None,
            older=util.normalize_url(item.older.url)
            if item.older
            else None,
            post_list=self.post_list,
        )



import xml.etree.ElementTree as etree
from collections.abc import Sequence
from logging import Logger
from pathlib import Path
from typing import Any, override

from l2m4m import LaTeX2MathMLExtension
from markdown import Markdown
from markdown.extensions.abbr import AbbrExtension
from markdown.extensions.attr_list import AttrListExtension
from markdown.extensions.codehilite import CodeHiliteExtension
from markdown.extensions.def_list import DefListExtension
from markdown.extensions.footnotes import FootnoteExtension
from markdown.extensions.md_in_html import MarkdownInHtmlExtension
from markdown.extensions.sane_lists import SaneListExtension
from markdown.extensions.smarty import SmartyExtension
from markdown.extensions.tables import TableExtension
from markdown.extensions.toc import TocExtension
from markdown.treeprocessors import Treeprocessor
from pygments.formatters.html import HtmlFormatter
from pymdownx.betterem import BetterEmExtension
from pymdownx.caret import InsertSupExtension
from pymdownx.escapeall import EscapeAllExtension
from pymdownx.inlinehilite import InlineHiliteExtension
from pymdownx.smartsymbols import SmartSymbolsExtension
from pymdownx.superfences import SuperFencesCodeExtension
from pymdownx.tilde import DeleteSubExtension

from zona import util
from zona.config import ZonaConfig
from zona.layout import Layout
from zona.log import get_logger
from zona.metadata import Metadata
from zona.models import Item


class ZonaImageTreeprocessor(Treeprocessor):
    """Implement Zona's image caption rendering."""

    def __init__(self, md: Markdown):
        super().__init__()
        self.md: Markdown = md
        self.logger: Logger = get_logger()

    @override
    def run(self, root: etree.Element):
        for parent in root.iter():
            for idx, child in enumerate(list(parent)):
                if (
                    child.tag == "p"
                    and len(child) == 1
                    and child[0].tag == "img"
                ):
                    img = child[0]
                    div = etree.Element(
                        "div", {"class": "image-container"}
                    )
                    div.append(img)
                    title = img.attrib.get("alt", "")
                    if title:
                        raw_caption = self.md.convert(title)
                        caption_html = raw_caption.strip()
                        if caption_html.startswith(
                            "<p>"
                        ) and caption_html.endswith("</p>"):
                            caption_html = caption_html[3:-4]
                        caption = etree.Element("small")
                        caption.text = ""  # should be rendered
                        caption_html_element = etree.fromstring(
                            f"<span>{caption_html}</span>"
                        )
                        caption.append(caption_html_element)
                        div.append(caption)
                    parent[idx] = div


class ZonaLinkTreeprocessor(Treeprocessor):
    def __init__(
        self,
        config: ZonaConfig | None,
        resolve: bool = False,
        source: Path | None = None,
        layout: Layout | None = None,
        item_map: dict[Path, Item] | None = None,
    ):
        super().__init__()
        self.resolve: bool = resolve
        self.logger: Logger = get_logger()
        if self.resolve:
            assert source is not None
            assert layout is not None
            assert item_map is not None
            self.source: Path = source.resolve()
            self.layout: Layout = layout
            self.item_map: dict[Path, Item] = item_map
            self.config: ZonaConfig | None = config

    @override
    def run(self, root: etree.Element):
        for element in root.iter("a"):
            href = element.get("href")
            if not href:
                continue
            if self.resolve:
                assert self.config
                cur = Path(href)
                _href = href
                same_file = False
                resolved = Path()
                # href starting with anchor reference the current file
                if href.startswith("#"):
                    same_file = True
                elif href.startswith("/"):
                    # resolve relative to content root
                    resolved = (
                        self.layout.content / cur.relative_to("/")
                    ).resolve()
                else:
                    # treat as relative link and try to resolve
                    resolved = (self.source.parent / cur).resolve()
                # check if the link is internal
                internal = same_file
                if not same_file:
                    for suffix in {".md", ".html"}:
                        if resolved.with_suffix(suffix).exists():
                            internal = True
                            resolved = resolved.with_suffix(suffix)
                            break
                # only substitute if link points to an actual file
                # that isn't the self file
                if not same_file and internal:
                    item = self.item_map.get(resolved)
                    if item:
                        href = util.normalize_url(item.url)
                        # don't sub if it's already correct lol
                        if _href != href:
                            element.set("href", href)
                            self.logger.debug(
                                f"Link in file {self.source}: {_href} resolved to {href}"
                            )
                    else:
                        self.logger.debug(
                            f"Warning: resolved path {resolved} not found in item map"
                        )
                # open link in new tab if not self-link
                elif (
                    self.config.markdown.links.external_new_tab
                    and not same_file
                ):
                    element.set("target", "_blank")


def get_formatter(config: ZonaConfig):
    c = config.markdown.syntax_highlighting
    formatter = HtmlFormatter(
        style=c.theme, nowrap=not c.wrap, nobackground=True
    )
    return formatter


def md_to_html(
    content: str,
    config: ZonaConfig | None,
    resolve_links: bool = False,
    source: Path | None = None,
    layout: Layout | None = None,
    item_map: dict[Path, Item] | None = None,
    metadata: Metadata | None = None,
) -> str:
    extensions: Sequence[Any] = [
        BetterEmExtension(),
        SuperFencesCodeExtension(
            disable_indented_code_blocks=True,
            css_class="codehilite",
        ),
        FootnoteExtension(),
        AttrListExtension(),
        DefListExtension(),
        TocExtension(
            anchorlink=True,
        ),
        TableExtension(),
        AbbrExtension(),
        SmartyExtension(),
        InsertSupExtension(),
        DeleteSubExtension(),
        SmartSymbolsExtension(),
        SaneListExtension(),
        MarkdownInHtmlExtension(),
        EscapeAllExtension(hardbreak=True),
    ]
    kwargs: dict[str, Any] = {
        "extensions": extensions,
        "tab_length": 2,
    }
    if metadata and metadata.math:
        kwargs["extensions"].append(LaTeX2MathMLExtension())
    if config:
        kwargs["extensions"].extend(
            [
                CodeHiliteExtension(
                    linenums=False,
                    noclasses=False,
                    pygments_style=config.markdown.syntax_highlighting.theme,
                ),
                InlineHiliteExtension(css_class="codehilite"),
            ]
        )
        kwargs["tab_length"] = config.markdown.tab_length
    md = Markdown(**kwargs)
    if resolve_links:
        if source is None or layout is None or item_map is None:
            raise TypeError(
                "md_to_html() missing source and ctx when resolve_links is true"
            )
        md.treeprocessors.register(
            item=ZonaLinkTreeprocessor(
                config, resolve_links, source, layout, item_map
            ),
            name="zona_links",
            priority=15,
        )
    md.treeprocessors.register(
        item=ZonaImageTreeprocessor(md),
        name="zona_images",
        priority=17,
    )
    return md.convert(content)


def get_style_defs(config: ZonaConfig) -> str:
    formatter = get_formatter(config)
    defs = formatter.get_style_defs(".codehilite")
    assert isinstance(defs, str)
    return defs



from pathlib import Path

from zona import util


def test_title(tmp_path: Path):
    a = tmp_path / "my-first-post.md"
    b = tmp_path / "Writing_emails_Post.md"
    assert util.filename_to_title(a) == "My First Post"
    assert util.filename_to_title(b) == "Writing Emails Post"


def test_get_resourcees():
    util.get_resources("templates")
    assert True
